import os
import sys
import time
import subprocess

autisto_dir = os.path.dirname(os.path.realpath(__file__))


class Platform:
    def __init__(self, name, service_name, service_file_dir, service_file_path):
        self.name = name
        self.service_name = service_name
        self.service_file_dir = service_file_dir
        self.service_file_path = service_file_path

    class NotImplemented(Exception):
        pass

    def service_active(self):
        raise self.NotImplemented

    def set_service(self):
        raise self.NotImplemented

    def remove_service(self):
        raise self.NotImplemented

    def start_service(self):
        raise self.NotImplemented

    def stop_service(self):
        raise self.NotImplemented


class Linux(Platform):
    def __init__(self):
        service_name = "autisto"
        service_file_dir = os.path.join("/etc/systemd/system")
        service_file_path = os.path.join(service_file_dir, f"{service_name}.service")

        super().__init__(
            "Linux",
            service_name,
            service_file_dir,
            service_file_path
        )

    def service_active(self, quiet=False):
        if subprocess.call(["systemctl", "is-active", "--quiet", self.service_name]) == 0:
            if not quiet:
                print("\033[92mService is active.\033[0m")
            return True
        else:
            if not quiet:
                print("\033[91mService is inactive.\033[0m")
            return False

    def set_service(self):
        self.remove_service()

        os.makedirs(self.service_file_dir, exist_ok=True)
        with open(self.service_file_path, "w") as file:
            file.write("[Unit]\n")
            file.write("Description=Autisto systemd service\n")
            file.write("After=mongod.service\n")
            file.write("\n[Service]\n")
            file.write("User=root\n")
            file.write("Type=simple\n")
            file.write(f"ExecStart={sys.executable} "
                       f"{os.path.join(autisto_dir, 'server.py')}\n")
            file.write("StandardOutput=file:/tmp/autisto.log\n")
            file.write("StandardError=file:/tmp/autisto.log\n")
            file.write("RestartSec=5\n")
            file.write("Restart=always\n")
            file.write("\n[Install]\n")
            file.write("WantedBy=multi-user.target\n")

        systemd_setup = [
            ["systemctl", "enable", self.service_name],
            ["systemctl", "daemon-reload"],
            ["systemctl", "start", self.service_name]
        ]
        success = [subprocess.call(cmd) == 0 for cmd in systemd_setup]

        if all(success):
            time.sleep(3)
            print("\n\033[92mAutisto systemd service enabled. It should be reported as active."
                  "\nClose status with 'q'.\033[0m")
            subprocess.run(["systemctl", "status", self.service_name])
        else:
            print("\n\033[91mAutisto systemd service set-up failed!\033[0m")

    def remove_service(self):
        subprocess.run(["systemctl", "stop", self.service_name])
        subprocess.run(["systemctl", "disable", self.service_name])
        subprocess.run(["systemctl", "daemon-reload"])
        subprocess.run(["systemctl", "reset-failed"])

    def start_service(self):
        subprocess.run(["systemctl", "start", self.service_name])

    def stop_service(self):
        if self.service_active(quiet=True):
            return subprocess.call(["systemctl", "stop", self.service_name]) == 0
        else:
            return False


class MacOS(Platform):
    def __init__(self):
        service_name = "com.autisto.run"
        service_file_dir = os.path.join(os.path.expanduser("~/Library/LaunchAgents"))
        service_file_path = os.path.join(service_file_dir, f"{service_name}.plist")

        super().__init__(
            "macOS",
            service_name,
            service_file_dir,
            service_file_path
        )

    def service_active(self, quiet=False):
        if self.service_name in subprocess.check_output(["launchctl", "list"]).decode("utf-8"):
            if not quiet:
                print("\033[92mService is active.\033[0m")
            return True
        else:
            if not quiet:
                print("\033[91mService is inactive.\033[0m")
            return False

    def set_service(self):
        self.stop_service()

        os.makedirs(self.service_file_dir, exist_ok=True)
        with open(self.service_file_path, "w") as file:
            file.write('<?xml version="1.0" encoding="UTF-8"?>\n')
            file.write('<!DOCTYPE plist PUBLIC "-//Apple//DTD PLIST 1.0//EN" '
                       '"http://www.apple.com/DTDs/PropertyList-1.0.dtd">\n')
            file.write('<plist version="1.0">\n')
            file.write('<dict>\n')
            file.write('    <key>Label</key>\n')
            file.write(f'    <string>{self.service_name}</string>\n')
            file.write('    <key>ProgramArguments</key>\n')
            file.write('    <array>\n')
            file.write(f'        <string>{sys.executable}</string>\n')
            file.write(f'        <string>{os.path.join(autisto_dir, "server.py")}</string>\n')
            file.write('    </array>\n')
            file.write('    <key>KeepAlive</key>\n')
            file.write('    <true/>\n')
            file.write('</dict>\n')
            file.write('</plist>\n')

        self.start_service()
        print("\n\033[92mAutisto launchd service enabled.\033[0m")
        time.sleep(1)
        self.service_active()

    def remove_service(self):
        subprocess.call(["launchctl", "unload", "-w", self.service_file_path])

    def start_service(self):
        subprocess.call(["launchctl", "load", "-w", self.service_file_path])

    def stop_service(self):
        if self.service_active(quiet=True):
            return subprocess.call(["launchctl", "unload", self.service_file_path]) == 0
        else:
            return False


def check_platform():
    if sys.platform in ["linux", "darwin"]:
        return sys.platform
    else:
        print("\n\033[91mAutisto is available only for Linux and macOS.\033[0m")
        sys.exit(0)


def get_platform():
    this_platform = check_platform()
    if this_platform == "linux":
        return Linux()
    elif this_platform == "darwin":
        return MacOS()



import csv
import urllib.error
import urllib.request
from datetime import datetime
from dateutil.relativedelta import relativedelta


class FinanceModule:
    url = ("https://stat.gov.pl/download/gfx/portalinformacyjny/pl/defaultstronaopisowa/4741/1/1/miesieczne_wskazniki_"
           "cen_towarow_i_uslug_konsumpcyjnych_od_1982_roku_5.csv")

    def __init__(self):
        self.error = None
        self._calc_accumulated_inflation(self._extract_month_over_month_inflation(self._load_inflation_data()))

    def _load_inflation_data(self):
        try:
            response = urllib.request.urlopen(FinanceModule.url)
            return csv.reader([line.decode("windows-1250") for line in response.readlines()], delimiter=";")
        except Exception as e:
            self.error = e
            return []

    def _extract_month_over_month_inflation(self, raw_inflation_data):
        month_over_month_inflation_data = {}
        for row in raw_inflation_data:
            if "Poprzedni miesiąc = 100" == row[2]:
                try:
                    inflation_rate = float(row[5].replace(",", ".")) / 100
                except ValueError:
                    continue
                try:
                    if int(row[3]) in month_over_month_inflation_data.keys():
                        month_over_month_inflation_data[int(row[3])][int(row[4])] = inflation_rate
                    else:
                        month_over_month_inflation_data[int(row[3])] = {
                            int(row[4]): float(row[5].replace(",", ".")) / 100}
                except Exception as e:
                    self.error = e
                    return {}

        two_months_ago = datetime.now() - relativedelta(months=2)
        if self.error is None:
            try:
                month_over_month_inflation_data[two_months_ago.year][two_months_ago.month]
            except KeyError:
                self.error = KeyError(
                    f"Error: inflation data for {two_months_ago.strftime('%B')} {two_months_ago.year} "
                    f"is missing in GUS .csv file available at {FinanceModule.url}")
        return month_over_month_inflation_data

    def _calc_accumulated_inflation(self, month_over_month_inflation_data):
        self._current_time = datetime.now()
        self._accumulated_inflation = {}
        accumulated_inflation = 1.
        for year in reversed(range(1982, self._current_time.year + 1)):
            self._accumulated_inflation[year] = {}
            for month in reversed(range(1, 13)):
                try:
                    accumulated_inflation *= month_over_month_inflation_data[year][month]
                except KeyError:
                    pass
                self._accumulated_inflation[year][month] = accumulated_inflation

    def calc_adjusted_price(self, original_price, purchase_date):
        try:
            adjusted_value = original_price * self._accumulated_inflation[
                purchase_date.year][purchase_date.month]
        except KeyError:
            adjusted_value = original_price * self._accumulated_inflation[1983][1]
        return adjusted_value

    def calc_adjusted_value_and_depreciation(self, document):
        total_value = 0.
        depreciation = 0.
        for i in range(document["quantity"]):
            purchase_date = datetime.strptime(document["dates_of_purchase"][i], "%d-%m-%Y")
            value = self.calc_adjusted_price(document["prices"][i], purchase_date)
            total_value += value
            relative_delta = relativedelta(self._current_time, purchase_date)
            months_passed = relative_delta.years * 12 + relative_delta.months
            depreciation_ratio = min(1., months_passed / document["life_expectancy_months"])
            depreciation += value * depreciation_ratio
        return total_value, depreciation
